/*
 * Copyright (c) 2007 Koji Suzuki
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 *  modification, are permitted provided that the following conditions
 *  are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR
 * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
 * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
/* rccal firmware
 * tiny44 --- PA7, OC0B	  ADC1,PA1, AREF
 * tiny45 --- PB4, OC1B   ADC3,PB3
 * tiny461 -- PB3, OC1B   ADC4,PA5, AREF
 * tiny26  -- PB3, OC1B   ADC4,PA5, AREF
 */

#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <avr/wdt.h>
#include <avr/sleep.h>

#ifndef ADCD
#define ADCD	ADC
#endif
#ifndef ADCSRA
#define ADCSRA	ADCSR
#endif

#include "i2cdev.h"

//#define NO_ADC
//#define TEST_PORT

#define bit_set(port,bit)       { port |= (1<<(bit));}
#define bit_clear(port,bit)     { port &= ~(1<<(bit));}

#if defined(__AVR_ATtiny461__) || defined(__AVR_ATtiny861__)
#define CHAN_V833	0x04	/* ADC4 PA5 */
#define CHAN_TEMP	0x1f
#define TGT_PORT	PORTB
#define TGT_DDR		DDRB
#define TGT_BIT		PB3
#elif defined(__AVR_ATtiny44__) || defined(__AVR_ATtiny84__)
#define CHAN_V833	0x01	/* ADC1 PA1 */
#define CHAN_TEMP	0x22
#define TGT_PORT	PORTA
#define TGT_DDR		DDRA
#define TGT_BIT		PA7
#elif defined(__AVR_ATtiny25__) || defined(__AVR_ATtiny45__)
#define CHAN_V833	0x03	/* ADC3 PB3 */
#define CHAN_TEMP	0x0f
#define TGT_PORT	PORTB
#define TGT_DDR		DDRB
#define TGT_BIT		PB4
#elif defined(__AVR_ATtiny26__) 
#define CHAN_V833	0x04	/* ADC4 PA5 */
#define TGT_PORT	PORTB
#define TGT_DDR		DDRB
#define TGT_BIT		PB3
#endif

static void set_clkpr(uint8_t n) {
#ifdef CLKPR
	cli();
	CLKPR = (1<<CLKPCE);
	CLKPR = n;
	sei();
#endif
}

static inline uint8_t adc_is_busy() {
#ifdef NO_ADC
	return 0;
#else
	return bit_is_set(ADCSRA,ADSC);
#endif
}

static uint16_t do_adc_native(uint8_t admux) {
#ifdef NO_ADC
	return 0;
#else
	uint16_t r1,r2,r3,tmp;
	ADMUX = admux;
	ADCSRA |= (1<<ADSC);
	while (adc_is_busy()) {
		;
	}
	r1 = ADCD;
	ADCSRA |= (1<<ADSC);
	while (adc_is_busy()) {
		;
	}
	r2 = ADCD;
	ADCSRA |= (1<<ADSC);
	while (adc_is_busy()) {
		;
	}
	r3 = ADCD;
	if (r2 > r1) {
		tmp = r1; r1 = r2; r2 =tmp;
	}
	if (r3 > r1) {
		tmp = r1; r1 = r3; r3 =tmp;
	}
	if (r3 > r2) {
		tmp = r2; r2 = r3; r3 =tmp;
	}
	return r2;
#endif
}

void init_adc() {
#ifdef NO_ADC
#else
	/* setup ADC @ 1/64 clock :125 kHz@8MHz */
	ADCSRA = (6<<ADPS0) | (1<<ADEN);
#ifdef SLEEP_ENABLE
	sleep_enable();
#endif
#if defined(__AVR_ATtiny461__) || defined(__AVR_ATtiny861__)
#elif defined(__AVR_ATtiny44__) || defined(__AVR_ATtiny84__)
#elif defined(__AVR_ATtiny25__) || defined(__AVR_ATtiny45__)
	MCUCR |= (1<<PUD);
	DIDR0 = (1<<ADC3D);
	DDRB &= ~(1<<PB3);
	PORTB &= ~(1<<PB3);
#endif
#endif
}

static inline uint16_t temp_val() {
#if defined(__AVR_ATtiny461__) || defined(__AVR_ATtiny861__)
	bit_clear(ADCSRB , REFS2); // 1.1V
	bit_set(ADCSRB , MUX5);
	return do_adc_native((2<<REFS0) | CHAN_TEMP);
#elif defined(__AVR_ATtiny44__) || defined(__AVR_ATtiny84__)
	return do_adc_native((2<<REFS0) | CHAN_TEMP);
#elif defined(__AVR_ATtiny25__) || defined(__AVR_ATtiny45__)
	//return do_adc_native((2<<REFS0) | CHAN_TEMP);
	return do_adc_native((2<<REFS0) | CHAN_TEMP | (1<<REFS2));
#elif defined(__AVR_ATtiny26__) 
	return 0;
#endif
}

static inline uint16_t v833_val(int8_t rtype) {
#if defined(__AVR_ATtiny461__) || defined(__AVR_ATtiny861__)
	bit_clear(ADCSRB , MUX5);
	if (rtype == 0) {
		bit_clear(ADCSRB , REFS2); // 1.1V
		return do_adc_native((2<<REFS0) | CHAN_V833);
	}
	if (rtype == 1) {
		bit_set(ADCSRB , REFS2); // 2.56 V
		return do_adc_native((2<<REFS0) | CHAN_V833);
	} 
	if (rtype == 2) { // ext 2.5 V
		return do_adc_native((1<<REFS0) | CHAN_V833);
	}
	return do_adc_native((0<<REFS0) | CHAN_V833);
#elif defined(__AVR_ATtiny44__) || defined(__AVR_ATtiny84__)
	if (rtype == 0) return do_adc_native((2<<REFS0) | CHAN_V833);
	if (rtype == 2) return do_adc_native((1<<REFS0) | CHAN_V833);
	return do_adc_native((0<<REFS0) | CHAN_V833);
#elif defined(__AVR_ATtiny25__) || defined(__AVR_ATtiny45__)
	if (rtype == 0) return do_adc_native((2<<REFS0) | CHAN_V833);
	if (rtype == 1) return do_adc_native((2<<REFS0) | CHAN_V833 | (1<<REFS2));
	return do_adc_native((0<<REFS0) | CHAN_V833);
#elif defined(__AVR_ATtiny26__) 
	if (rtype == 1) return do_adc_native((2<<REFS0) | CHAN_V833);
	if (rtype == 2) return do_adc_native((1<<REFS0) | CHAN_V833);
	return do_adc_native((0<<REFS0) | CHAN_V833);
#endif
}

SIGNAL(SIG_ADC) {
}


uint8_t myaddr;

void set_osc(uint8_t target_osccal) {
	while (OSCCAL != target_osccal) {
		if (OSCCAL < target_osccal) OSCCAL++;
		else OSCCAL--;
	}
}

//static inline 
void init_pwm() {
	bit_set(TGT_DDR, TGT_BIT);

#ifndef TEST_PORT
#if defined(__AVR_ATtiny461__) || defined(__AVR_ATtiny861__)
	OCR1B = 0x1;
	OCR1C = 0x3; // top 
	TCCR1A = (1<<COM1B0) | (1<<PWM1B); // toggle , enable OC1B
	TCCR1D = (0<<WGM10); // fast pwm mode
	TCCR1B = 1; // start @ 1/1
#elif defined(__AVR_ATtiny44__) || defined(__AVR_ATtiny84__)
#if 1
	OCR0B = 0x1;
	OCR0A = 0x3; // top
	TCCR0A = (2<<COM0B0) | (3<<WGM00);
	TCCR0B = (1<<WGM02) | 1; // start @ 1/1 , ctc pwm mode (TOP=OCR0A)
#else
	OCR0A = 0x2;
	OCR0B = 0x1;
	TCCR0A = (1<<COM0B0) | (2<<WGM00); // toggle, TOP=OCR0A
	TCCR0B = (0<<WGM02) | 1; // start @ 1/1
#endif
#elif defined(__AVR_ATtiny25__) || defined(__AVR_ATtiny45__)
#if 0
	bit_set(PLLCSR, PLLE);
	while (bit_is_clear(PLLCSR, PLOCK))
		;
	bit_set(PLLCSR, PCKE);
#endif
	OCR1B = 0x2; // COR1B >= 2
	OCR1C = 0x3; // top
	GTCCR = (2<<COM1B0) | (1<<PWM1B);
	TCCR1 = (1<<CTC1) | 1; // start @ 1/1 , TOP=OCR0C
	TCCR1 |= (2<<COM1A0) | (1<<PWM1A); // XXX ??
#elif defined(__AVR_ATtiny26__) 
	OCR1B = 0x2;
	OCR1C = 0x3; // top 
	TCCR1A = (2<<COM1B0) | (1<<PWM1B); // toggle , enable OC1B
	TCCR1B = (1<<CTC1) | 1; // start @ 1/1
#endif
#endif
}

uint8_t cnt;

int main() {
	uint8_t ch;

#ifdef CLKPR
	/* set system_clock pre scaler 1/8 => 1/1 */
	CLKPR = (1<<CLKPCE);
	CLKPR = 0;
#endif


	myaddr = 0x10;
	i2cdev_init( myaddr );
	init_pwm();
	init_adc();

	sei();

	for(;;) {
		if( i2cdev_can_getc(1) ) {
			ch = i2cdev_getc();
			if (ch == 'S') {
				i2cdev_putc('R');
				i2cdev_putc('C');
				i2cdev_putc('C');
				i2cdev_putc('A');
				i2cdev_putc('L');
			} else if (ch == 'O' && i2cdev_can_getc(1)) {
				set_osc(i2cdev_getc());
				i2cdev_putc(OSCCAL);
			} else if (ch == 'o') {
				i2cdev_putc(OSCCAL);
			} else if (ch == 't') {
				uint16_t temp;
				temp = temp_val();
				i2cdev_putc((uint8_t)temp);
				i2cdev_putc((uint8_t)(temp>>8));
			} else if (ch == 'v' && i2cdev_can_getc(1)) {
				uint16_t v833;
				v833 = v833_val(i2cdev_getc());
				i2cdev_putc((uint8_t)v833);
				i2cdev_putc((uint8_t)(v833>>8));
			} else if ((ch == 'c') && i2cdev_can_getc(1)) {
				set_clkpr(i2cdev_getc());
			}
			i2cdev_flush();
		}
		cnt++;
#ifdef TEST_PORT
		if (cnt &1) {
			bit_set(TGT_PORT, TGT_BIT);
		} else {
			bit_clear(TGT_PORT, TGT_BIT);
		}
#endif
	}
}
